@@ -6,22 +6,22 @@ import re
 import shlex
 import asyncore
 import sys
-from py65.mpu6502 import MPU
+from py65.devices.mpu6502 import MPU as NMOS6502
+from py65.devices.mpu65c02 import MPU as CMOS65C02
 from py65.disassembler import Disassembler
 from py65.assembler import Assembler
-from py65.util import itoa, AddressParser, getch
+from py65.utils.addressing import AddressParser
+from py65.utils.console import getch
+from py65.utils.conversions import itoa
 from py65.memory import ObservableMemory
 
 class Monitor(cmd.Cmd):
 
-    def __init__(self, options={}, completekey='tab', stdin=None, stdout=None):
-        self.options = options
-        self._mpu = MPU()
-        self._install_mpu_observers()
+    def __init__(self, mpu_type=NMOS6502, completekey='tab', stdin=None, stdout=None):
+        self._reset(mpu_type)
+        self._width = 78
         self._update_prompt()
-        self._address_parser = AddressParser()
-        self._disassembler = Disassembler(self._mpu, self._address_parser)
-        self._assembler = Assembler(self._address_parser)
+        self._add_shortcuts()
         cmd.Cmd.__init__(self, completekey, stdin, stdout)
 
     def onecmd(self, line):
@@ -40,12 +40,39 @@ class Monitor(cmd.Cmd):
         self._update_prompt()
         return result
 
+    def _reset(self, mpu_type):
+        self._mpu = mpu_type()
+        self._install_mpu_observers()
+        self._address_parser = AddressParser()
+        self._disassembler = Disassembler(self._mpu, self._address_parser)
+        self._assembler = Assembler(self._mpu, self._address_parser)
+
+    def _add_shortcuts(self):
+        self._shortcuts = {'~':   'tilde',
+                           '?':   'help',
+                           'al':  'add_label',
+                           'd':   'disassemble',
+                           'dl':  'delete_label',
+                           'f':   'fill',
+                           'g':   'goto',
+                           'l':   'load',
+                           'll':  'load_labels',
+                           'm':   'mem',
+                           'r':   'registers',
+                           'ret': 'return',
+                           'rad': 'radix',
+                           'shl': 'show_labels',
+                           'x':   'quit',
+                           'z':   'step'}
+
     def _preprocess_line(self, line):
-        # tilde command
-        matches = re.match('^~\s*', line)
-        if matches:
-            start, end = matches.span()
-            line = "tilde %s" % line[end:]
+        # command shortcuts
+        for shortcut, command in self._shortcuts.iteritems():
+            pattern = '^%s\s+' % re.escape(shortcut)
+            matches = re.match(pattern, line)
+            if matches:
+                start, end = matches.span()
+                line = "%s %s" % (command, line[end:])
 
         # line comments
         quoted = False
@@ -63,12 +90,12 @@ class Monitor(cmd.Cmd):
             self.stdout.write(chr(value))
             self.stdout.flush()
 
-        def getc(operation, address, value):
+        def getc(operation, address):
             return getch(self.stdin)
 
         m = ObservableMemory()
-        m.subscribe(m.WRITE, [0xF001], putc)
-        m.subscribe(m.READ,  [0xF004], getc)
+        m.subscribe_to_write([0xF001], putc)
+        m.subscribe_to_read([0xF004], getc)
         
         self._mpu.memory = m
 
@@ -93,8 +120,32 @@ class Monitor(cmd.Cmd):
         self._output("reset\t\tReset the microprocessor")
 
     def do_reset(self, args):
-        self._mpu = MPU()
-        self._install_mpu_observers()
+        klass = self._mpu.__class__
+        self._reset(mpu_type=klass)
+
+    def do_mpu(self, args):                                         
+        mpus = {'6502': NMOS6502, '65C02': CMOS65C02}
+        
+        def available_mpus():
+            mpu_list = ', '.join(mpus.keys())
+            self._output("Available MPUs: %s" % mpu_list)            
+        
+        if args == '':                      
+            self._output("Current MPU is %s" % self._mpu.name)
+            available_mpus()
+        else:
+            requested = args.upper()
+            new_mpu = mpus.get(requested, None)
+            if new_mpu is None:
+                self._output("Unknown MPU: %s" % args)
+                available_mpus()
+            else:
+                self._reset(new_mpu)
+                self._output("Reset with new MPU %s" % self._mpu.name)
+
+    def help_mpu(self):
+        self._output("mpu\t\tPrint available microprocessors.")
+        self._output("mpu <type>\tSelect a new microprocessor.")
         
     def do_EOF(self, args):
         self._output('')
@@ -104,7 +155,7 @@ class Monitor(cmd.Cmd):
         self._output("To quit, type ^D or use the quit command.")
 
     def do_quit(self, args):
-        return self.do_EOF
+        return self.do_EOF(args)
     
     def help_quit(self):
         return self.help_EOF()
@@ -115,11 +166,15 @@ class Monitor(cmd.Cmd):
             return self.help_assemble()
         
         start, statement = split
-        start = self._address_parser.number(start)
+        try:
+            start = self._address_parser.number(start)
+        except KeyError:
+            self._output("Bad label: %s" % start)
+            return
 
         bytes = self._assembler.assemble(statement)
         if bytes is None:
-            self._output("Statement could not be assembled.")
+            self._output("Assemble failed: %s" % statement)
         else:
             end = start + len(bytes)
             self._mpu.memory[start:end] = bytes
@@ -213,7 +268,7 @@ class Monitor(cmd.Cmd):
         self._output("~ <number>")
         self._output("Display the specified number in decimal, hex, octal, and binary.")
     
-    def do_tilde(self, args):
+    def do_tilde(self, args):   
         try:
             num = self._address_parser.number(args)
         except ValueError:
@@ -247,8 +302,8 @@ class Monitor(cmd.Cmd):
                     if len(register) == 1:
                         intval &= 0xFF
                     setattr(self._mpu, register, intval)
-                except ValueError:
-                    self._output("Syntax error: %s" % value)
+                except KeyError, why:
+                    self._output(why[0])
     
     def help_cd(self, args):
         self._output("cd <directory>")
@@ -298,10 +353,10 @@ class Monitor(cmd.Cmd):
         self._fill(start, start, map(ord, bytes))
     
     def help_fill(self):
-        self.output("fill <address_range> <data_list>")
-        self.output("Fill memory in the specified address range with the data in")
-        self.output("<data_list>.  If the size of the address range is greater")
-        self.output("than the size of the data_list, the data_list is repeated.")
+        self._output("fill <address_range> <data_list>")
+        self._output("Fill memory in the specified address range with the data in")
+        self._output("<data_list>.  If the size of the address range is greater")
+        self._output("than the size of the data_list, the data_list is repeated.")
     
     def do_fill(self, args):
         split = shlex.split(args)
@@ -341,10 +396,17 @@ class Monitor(cmd.Cmd):
     def do_mem(self, args):
         start, end = self._address_parser.range(args)
 
-        out = itoa(start, 16).zfill(4) + ":  "
-        for byte in self._mpu.memory[start:end+1]:
-            out += "%02x  " % byte
-        self._output(out)
+        line = "%04x:" % start
+        for address in range(start, end+1):
+            byte = self._mpu.memory[address]
+            more = "  %02x" % byte                       
+            
+            exceeded = len(line) + len(more) > self._width
+            if exceeded:
+                self._output(line)
+                line = "%04x:" % address
+            line += more
+        self._output(line)
 
     def help_add_label(self):
         self._output("add_label <address> <label>")
@@ -379,15 +441,34 @@ class Monitor(cmd.Cmd):
         self._output("Remove the specified label from the label tables.")
 
     def do_delete_label(self, args):
+        if args == '':
+            return self.help_delete_label()
+        
         try:
             del self._address_parser.labels[args]
         except KeyError:
             pass
 
+    def do_width(self, args):
+        if args != '':
+            try:
+                new_width = int(args)
+                if new_width >= 10:
+                    self._width = new_width
+                else:
+                    self._output("Minimum terminal width is 10")
+            except ValueError:
+                self._output("Illegal width: %s" % args)
 
+        self._output("Terminal width is %d" % self._width)
 
-def main(args=None, options=None):
-    c = Monitor(options)
+    def help_width(self):
+        self._output("width <columns>")
+        self._output("Set the width used by some commands to wrap output.")
+        self._output("With no argument, the current width is printed.")
+
+def main(args=None):
+    c = Monitor()
 
     try:
         import readline
